@@ -18,7 +18,8 @@ use super::{
     HandshakeRole, MidHandshake, ProcessingResult,
 };
 use crate::{
-    error::{Error, ProtocolError, Result, UrlError},
+    error::{Error, ProtocolError, Result, SubProtocolError, UrlError},
+    handshake::version_as_str,
     protocol::{Role, WebSocket, WebSocketConfig},
 };
 
@@ -26,7 +27,7 @@ use crate::{
 pub type Request = HttpRequest<()>;
 
 /// Client response type.
-pub type Response = HttpResponse<()>;
+pub type Response = HttpResponse<Option<Vec<u8>>>;
 
 /// Client handshake role.
 #[derive(Debug)]
@@ -54,6 +55,8 @@ impl<S: Read + Write> ClientHandshake<S> {
         // Check the URI scheme: only ws or wss are supported
         let _ = crate::client::uri_mode(request.uri())?;
 
+        let subprotocols = extract_subprotocols_from_request(&request)?;
+
         // Convert and verify the `http::Request` and turn it into the request as per RFC.
         // Also extract the key from it (it must be present in a correct request).
         let (request, key) = generate_request(request)?;
@@ -62,7 +65,11 @@ impl<S: Read + Write> ClientHandshake<S> {
 
         let client = {
             let accept_key = derive_accept_key(key.as_ref());
-            ClientHandshake { verify_data: VerifyData { accept_key }, config, _marker: PhantomData }
+            ClientHandshake {
+                verify_data: VerifyData { accept_key, subprotocols },
+                config,
+                _marker: PhantomData,
+            }
         };
 
         trace!("Client handshake initiated.");
@@ -83,7 +90,15 @@ impl<S: Read + Write> HandshakeRole for ClientHandshake<S> {
                 ProcessingResult::Continue(HandshakeMachine::start_read(stream))
             }
             StageResult::DoneReading { stream, result, tail } => {
-                let result = self.verify_data.verify_response(result)?;
+                let result = match self.verify_data.verify_response(result) {
+                    Ok(r) => r,
+                    Err(Error::Http(mut e)) => {
+                        *e.body_mut() = Some(tail);
+                        return Err(Error::Http(e));
+                    }
+                    Err(e) => return Err(e),
+                };
+
                 debug!("Client handshake done.");
                 let websocket =
                     WebSocket::from_partially_read(stream, tail, Role::Client, self.config);
@@ -94,13 +109,13 @@ impl<S: Read + Write> HandshakeRole for ClientHandshake<S> {
 }
 
 /// Verifies and generates a client WebSocket request from the original request and extracts a WebSocket key from it.
-fn generate_request(mut request: Request) -> Result<(Vec<u8>, String)> {
+pub fn generate_request(mut request: Request) -> Result<(Vec<u8>, String)> {
     let mut req = Vec::new();
     write!(
         req,
-        "GET {path} {version:?}\r\n",
+        "GET {path} {version}\r\n",
         path = request.uri().path_and_query().ok_or(Error::Url(UrlError::NoPathOrQuery))?.as_str(),
-        version = request.version()
+        version = version_as_str(request.version())?,
     )
     .unwrap();
 
@@ -115,7 +130,7 @@ fn generate_request(mut request: Request) -> Result<(Vec<u8>, String)> {
         .get(KEY_HEADERNAME)
         .ok_or_else(|| {
             Error::Protocol(ProtocolError::InvalidHeader(
-                HeaderName::from_bytes(KEY_HEADERNAME.as_bytes()).unwrap(),
+                HeaderName::from_bytes(KEY_HEADERNAME.as_bytes()).unwrap().into(),
             ))
         })?
         .to_str()?
@@ -129,27 +144,35 @@ fn generate_request(mut request: Request) -> Result<(Vec<u8>, String)> {
     //
     // See similar problem in `hyper`: https://github.com/hyperium/hyper/issues/1492
     let headers = request.headers_mut();
-    for header in WEBSOCKET_HEADERS {
+    for &header in &WEBSOCKET_HEADERS {
         let value = headers.remove(header).ok_or_else(|| {
             Error::Protocol(ProtocolError::InvalidHeader(
-                HeaderName::from_bytes(header.as_bytes()).unwrap(),
+                HeaderName::from_bytes(header.as_bytes()).unwrap().into(),
             ))
         })?;
-        write!(req, "{header}: {value}\r\n", header = header, value = value.to_str()?).unwrap();
+        write!(
+            req,
+            "{header}: {value}\r\n",
+            header = header,
+            value = value.to_str().map_err(|err| {
+                Error::Utf8(format!("{err} for header name '{header}' with value: {value:?}"))
+            })?
+        )
+        .unwrap();
     }
 
     // Now we must ensure that the headers that we've written once are not anymore present in the map.
     // If they do, then the request is invalid (some headers are duplicated there for some reason).
-    let insensitive: Vec<String> =
-        WEBSOCKET_HEADERS.iter().map(|h| h.to_ascii_lowercase()).collect();
+    let websocket_headers_contains =
+        |name| WEBSOCKET_HEADERS.iter().any(|h| h.eq_ignore_ascii_case(name));
+
     for (k, v) in headers {
         let mut name = k.as_str();
 
         // We have already written the necessary headers once (above) and removed them from the map.
         // If we encounter them again, then the request is considered invalid and error is returned.
-        // Note that we can't use `.contains()`, since `&str` does not coerce to `&String` in Rust.
-        if insensitive.iter().any(|x| x == name) {
-            return Err(Error::Protocol(ProtocolError::InvalidHeader(k.clone())));
+        if websocket_headers_contains(name) {
+            return Err(Error::Protocol(ProtocolError::InvalidHeader(k.clone().into())));
         }
 
         // Relates to the issue of some servers treating headers in a case-sensitive way, please see:
@@ -158,7 +181,19 @@ fn generate_request(mut request: Request) -> Result<(Vec<u8>, String)> {
             name = "Sec-WebSocket-Protocol";
         }
 
-        writeln!(req, "{}: {}\r", name, v.to_str()?).unwrap();
+        if name == "origin" {
+            name = "Origin";
+        }
+
+        writeln!(
+            req,
+            "{}: {}\r",
+            name,
+            v.to_str().map_err(|err| {
+                Error::Utf8(format!("{err} for header name '{name}' with value: {v:?}"))
+            })?
+        )
+        .unwrap();
     }
 
     writeln!(req, "\r").unwrap();
@@ -166,11 +201,22 @@ fn generate_request(mut request: Request) -> Result<(Vec<u8>, String)> {
     Ok((req, key))
 }
 
+fn extract_subprotocols_from_request(request: &Request) -> Result<Option<Vec<String>>> {
+    if let Some(subprotocols) = request.headers().get("Sec-WebSocket-Protocol") {
+        Ok(Some(subprotocols.to_str()?.split(',').map(|s| s.trim().to_string()).collect()))
+    } else {
+        Ok(None)
+    }
+}
+
 /// Information for handshake verification.
 #[derive(Debug)]
 struct VerifyData {
     /// Accepted server key.
     accept_key: String,
+
+    /// Accepted subprotocols
+    subprotocols: Option<Vec<String>>,
 }
 
 impl VerifyData {
@@ -178,7 +224,7 @@ impl VerifyData {
         // 1. If the status code received from the server is not 101, the
         // client handles the response per HTTP [RFC2616] procedures. (RFC 6455)
         if response.status() != StatusCode::SWITCHING_PROTOCOLS {
-            return Err(Error::Http(response.map(|_| None)));
+            return Err(Error::Http(response.into()));
         }
 
         let headers = response.headers();
@@ -226,7 +272,27 @@ impl VerifyData {
         // not present in the client's handshake (the server has indicated a
         // subprotocol not requested by the client), the client MUST _Fail
         // the WebSocket Connection_. (RFC 6455)
-        // TODO
+        if headers.get("Sec-WebSocket-Protocol").is_none() && self.subprotocols.is_some() {
+            return Err(Error::Protocol(ProtocolError::SecWebSocketSubProtocolError(
+                SubProtocolError::NoSubProtocol,
+            )));
+        }
+
+        if headers.get("Sec-WebSocket-Protocol").is_some() && self.subprotocols.is_none() {
+            return Err(Error::Protocol(ProtocolError::SecWebSocketSubProtocolError(
+                SubProtocolError::ServerSentSubProtocolNoneRequested,
+            )));
+        }
+
+        if let Some(returned_subprotocol) = headers.get("Sec-WebSocket-Protocol") {
+            if let Some(accepted_subprotocols) = &self.subprotocols {
+                if !accepted_subprotocols.contains(&returned_subprotocol.to_str()?.to_string()) {
+                    return Err(Error::Protocol(ProtocolError::SecWebSocketSubProtocolError(
+                        SubProtocolError::InvalidSubProtocol,
+                    )));
+                }
+            }
+        }
 
         Ok(response)
     }
@@ -246,12 +312,12 @@ impl TryParse for Response {
 impl<'h, 'b: 'h> FromHttparse<httparse::Response<'h, 'b>> for Response {
     fn from_httparse(raw: httparse::Response<'h, 'b>) -> Result<Self> {
         if raw.version.expect("Bug: no HTTP version") < /*1.*/1 {
-            return Err(Error::Protocol(ProtocolError::WrongHttpMethod));
+            return Err(Error::Protocol(ProtocolError::WrongHttpVersion));
         }
 
         let headers = HeaderMap::from_httparse(raw.headers)?;
 
-        let mut response = Response::new(());
+        let mut response = Response::new(None);
         *response.status_mut() = StatusCode::from_u16(raw.code.expect("Bug: no HTTP status code"))?;
         *response.headers_mut() = headers;
         // TODO: httparse only supports HTTP 0.9/1.0/1.1 but not HTTP 2.0
@@ -267,7 +333,7 @@ pub fn generate_key() -> String {
     // a base64-encoded (see Section 4 of [RFC4648]) value that,
     // when decoded, is 16 bytes in length (RFC 6455)
     let r: [u8; 16] = rand::random();
-    base64::encode(&r)
+    data_encoding::BASE64.encode(&r)
 }
 
 #[cfg(test)]
@@ -278,9 +344,9 @@ mod tests {
     #[test]
     fn random_keys() {
         let k1 = generate_key();
-        println!("Generated random key 1: {}", k1);
+        println!("Generated random key 1: {k1}");
         let k2 = generate_key();
-        println!("Generated random key 2: {}", k2);
+        println!("Generated random key 2: {k2}");
         assert_ne!(k1, k2);
         assert_eq!(k1.len(), k2.len());
         assert_eq!(k1.len(), 24);
@@ -300,9 +366,7 @@ mod tests {
             Upgrade: websocket\r\n\
             Sec-WebSocket-Version: 13\r\n\
             Sec-WebSocket-Key: {key}\r\n\
-            \r\n",
-            host = host,
-            key = key
+            \r\n"
         )
         .into_bytes()
     }
